/*
	MSmooth plugin for Avisynth -- performs detail-preserving smoothing.

	Copyright (C) 2003 Donald A. Graft

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "internal.h"
#include "windows.h"
#include "stdio.h"
#include "msmooth.h"

void DrawStringRGB(PVideoFrame &dst, int x, int y, int h, const char *s); 

PVideoFrame __stdcall Msmooth::GetFrame(int n, IScriptEnvironment* env)
{
	if (vi.IsRGB()) return(GetFrameRGB(n, env));
	else if (vi.IsYV12()) return(GetFrameYV12(n, env));
	/* Avoid compiler warning. */
	return 0;
}

PVideoFrame __stdcall Msmooth::GetFrameRGB(int n, IScriptEnvironment* env)
{

    PVideoFrame src = child->GetFrame(n, env);
    PVideoFrame blur = env->NewVideoFrame(vi);
    PVideoFrame work = env->NewVideoFrame(vi);
    PVideoFrame mask = env->NewVideoFrame(vi);
    PVideoFrame final = env->NewVideoFrame(vi);
    PVideoFrame final2 = env->NewVideoFrame(vi);
	PVideoFrame deliver;

	const unsigned char *srcp = src->GetReadPtr();
	const unsigned char *srcp_saved = srcp;
    int src_pitch = src->GetPitch();

    unsigned char *blurp = blur->GetWritePtr();
	unsigned char *blurp_saved = blurp;
    int blur_pitch = blur->GetPitch();

    unsigned char *workp = work->GetWritePtr();
	unsigned char *workp_saved = workp;

    unsigned char *maskp = mask->GetWritePtr();
	unsigned char *maskp_saved = maskp;

    unsigned char *finalp = final->GetWritePtr();
	unsigned char *finalp_saved = finalp;

    unsigned char *finalp2 = final2->GetWritePtr();
	unsigned char *finalp2_saved = finalp2;

    int row_size = blur->GetRowSize();
    int height = blur->GetHeight();
	int y, reps;

	/* Create the detail mask. */
	EdgeMaskRGB(srcp, blurp, workp, maskp, row_size, height, src_pitch, blur_pitch);

	if (showmask == true)
	{
		if (show == true)
		{
			char buf[80];
			env->MakeWritable(&mask);
			sprintf(buf, VERSION);
			DrawStringRGB(mask, 0, 0, height, buf);
			sprintf(buf, COPYRIGHT);
			DrawStringRGB(mask, 0, 1, height, buf);
		}
		return mask;
	}

	/* Fix up output frame borders. */
	srcp = srcp_saved;
	finalp = finalp_saved;
	finalp2 = finalp2_saved;
	memcpy(finalp, srcp, row_size);
	memcpy(finalp + (height-1)*blur_pitch, srcp + (height-1)*blur_pitch, row_size);
	memcpy(finalp2, srcp, row_size);
	memcpy(finalp2 + (height-1)*blur_pitch, srcp + (height-1)*blur_pitch, row_size);
	for (y = 0; y < height; y++)
	{
		*((unsigned int *)(&finalp[0])) = *((unsigned int *)(&srcp[0]));
		*((unsigned int *)(&finalp2[0])) = *((unsigned int *)(&srcp[0]));
		*((unsigned int *)(&finalp[row_size-4])) = *((unsigned int *)(&srcp[row_size-4]));
		*((unsigned int *)(&finalp2[row_size-4])) = *((unsigned int *)(&srcp[row_size-4]));
		srcp += src_pitch;
		finalp += blur_pitch;
		finalp2 += blur_pitch;
	}

	/* Masked smoothing. */
	reps = 0;
	SmoothingPassRGB(srcp_saved, maskp_saved, workp_saved, finalp_saved, row_size, height, src_pitch, blur_pitch);
	if (++reps >= strength)
	{
		deliver = final;
		goto done;
	}
	while (1)
	{
		SmoothingPassRGB(finalp_saved, maskp_saved, workp_saved, finalp2_saved, row_size, height, blur_pitch, blur_pitch);
		if (++reps >= strength)
		{
			deliver = final2;
			goto done;
		}
		SmoothingPassRGB(finalp2_saved, maskp_saved, workp_saved, finalp_saved, row_size, height, blur_pitch, blur_pitch);
		if (++reps >= strength)
		{
			deliver = final;
			goto done;
		}
	}
	
done:
	if (show == true)
	{
		char buf[80];
		env->MakeWritable(&deliver);
		sprintf(buf, VERSION);
		DrawStringRGB(deliver, 0, 0, height, buf);
		sprintf(buf, COPYRIGHT);
		DrawStringRGB(deliver, 0, 1, height, buf);
	}
	return(deliver);

}

void __stdcall Msmooth::EdgeMaskRGB(const unsigned char *srcp, unsigned char *blurp, unsigned char *workp, unsigned char *maskp,
                        int row_size, int height, int src_pitch, int blur_pitch)
{
	const unsigned char *srcp_saved = srcp;
	const unsigned char *srcpp;
	const unsigned char *srcpn;

	unsigned char *blurp_saved = blurp;
	const unsigned char *blurpn;

	unsigned char *workp_saved = workp;
	unsigned char *workpp;
	unsigned char *workpn;

	unsigned char *maskp_saved = maskp;

	int x, y;
	int r1, r2, r3, r4, g1, g2, g3, g4, b1, b2, b3, b4;

	/* Blur the source image prior to detail detection. */
	srcpp = srcp_saved;
	srcp  = srcpp + src_pitch;
	srcpn = srcp + src_pitch;
	workp = workp_saved + blur_pitch;
	memcpy(workp_saved, srcp_saved, row_size);
	memcpy(workp_saved + (height-1)*blur_pitch, srcp_saved + (height-1)*src_pitch, row_size);
	for (y = 1; y < height - 1; y++)
	{
		workp[0] = srcp[0];
		workp[1] = srcp[1];
		workp[2] = srcp[2];
		workp[row_size-4] = srcp[row_size-4];
		workp[row_size-3] = srcp[row_size-3];
		workp[row_size-2] = srcp[row_size-2];
        for (x = 4; x < row_size - 4; x++)
		{
			workp[x] = (srcpp[x] + srcp[x] + srcpn[x]) / 3;
		}
        srcpp += src_pitch;
        srcp += src_pitch;
        srcpn += src_pitch;
        workp += blur_pitch;
    }
	srcp  = srcp_saved + src_pitch;
	workpp = workp_saved;
	workp  = workpp + blur_pitch;
	workpn = workp + blur_pitch;
	blurp  = blurp_saved + blur_pitch;
	memcpy(blurp_saved, srcp_saved, row_size);
	memcpy(blurp_saved + (height-1)*blur_pitch, srcp_saved + (height-1)*src_pitch, row_size);
	for (y = 1; y < height - 1; y++)
	{
		blurp[0] = srcp[0];
		blurp[1] = srcp[1];
		blurp[2] = srcp[2];
		blurp[row_size-4] = srcp[row_size-4];
		blurp[row_size-3] = srcp[row_size-3];
		blurp[row_size-2] = srcp[row_size-2];
        for (x = 4; x < row_size - 4; x++)
		{
			blurp[x] = (workp[x-4] + workp[x] + workp[x+4]) / 3;
		}
        srcp += src_pitch;
        workpp += blur_pitch;
        workp += blur_pitch;
        workpn += blur_pitch;
        blurp += blur_pitch;
    }

	/* Diagonal detail detection. */
	blurp = (unsigned char *) workp_saved;
	blurpn = blurp + blur_pitch;
	maskp = maskp_saved;
	b1 = blurp[0];
	g1 = blurp[1];
	r1 = blurp[2];
	b3 = blurpn[0];
	g3 = blurpn[1];
	r3 = blurpn[2];
	for (y = 0; y < height - 1; y++) 
	{
		for (x = 0; x < row_size - 4; x+=4)
		{
			b2 = blurp[x+4];
			g2 = blurp[x+5];
			r2 = blurp[x+6];
			b4 = blurpn[x+4];
			g4 = blurpn[x+5];
			r4 = blurpn[x+6];
			if ((abs(b1 - b4) >= threshold) || (abs(g1 - g4) >= threshold) || (abs(r1 - r4) >= threshold) ||
				(abs(b2 - b3) >= threshold) || (abs(g2 - g3) >= threshold) || (abs(g2 - g3) >= threshold))
			{
				*((unsigned int *)(&maskp[x])) |= 0xffffffff;
			}
			else
			{
				*((unsigned int *)(&maskp[x])) = 0x0;
			}
			b1 = b2; b3 = b4;
			g1 = g2; g3 = g4;
			r1 = r2; r3 = r4;
		}
		maskp += blur_pitch;
		blurp += blur_pitch;
		blurpn += blur_pitch;
	}

	/* If set for high quality, also do horizontal and vertical
	   detail detection. */
	if (highq == true)
	{

		/* Msmooth convolve vertical. */
		for (x = 0; x < row_size; x+=4)
		{
 			blurp = blurp_saved;
			blurpn = blurp + blur_pitch;
			maskp = maskp_saved;
			b1 = blurp[x];
			g1 = blurp[x+1];
			r1 = blurp[x+2];
			for (y = 0; y < height - 1; y++)
			{
				b2 = blurpn[x];
				g2 = blurpn[x+1];
				r2 = blurpn[x+2];
				if (abs(b1 - b2) >= threshold || abs(g1 - g2) >= threshold || abs(r1 - r2) >= threshold)
				{
					*((unsigned int *)(&maskp[x])) |= 0xffffffff;
				}
				b1 = b2;
				g1 = g2;
				r1 = r2;
				maskp += blur_pitch;
				blurp += blur_pitch;
				blurpn += blur_pitch;
			}
		}

		/* Msmooth convolve horizontal. */
		blurp = blurp_saved;
		maskp = maskp_saved;
		for (y = 0; y < height; y++)
		{
			b1 = blurp[0];
			g1 = blurp[1];
			r1 = blurp[2];
			for (x = 0; x < row_size - 4; x+=4)
			{
				b2 = blurp[x+4];
				g2 = blurp[x+5];
				r2 = blurp[x+6];
				if (abs(b1 - b2) >= threshold || abs(g1 - g2) >= threshold || abs(r1 - r2) >= threshold)
				{
					*((unsigned int *)(&maskp[x])) = 0xffffffff;
				}
				b1 = b2;
				g1 = g2;
				r1 = r2;
			}
			maskp += blur_pitch;
			blurp += blur_pitch;
		}
	}

	/* Fix up detail map borders. */
	maskp = maskp_saved;
	memset(maskp, 0xff, row_size);
	memset(maskp + (height-1)*blur_pitch, 0xff, row_size);
	for (y = 0; y < height; y++)
	{
		*((unsigned int *)(&maskp[0])) = 0xffffffff;
		*((unsigned int *)(&maskp[row_size-4])) = 0xffffffff;
		maskp += blur_pitch;
	}
}

void __stdcall Msmooth::SmoothingPassRGB(const unsigned char *srcp, unsigned char *maskp,
									  unsigned char *workp, unsigned char *finalp,
									  int row_size, int height, int spitch, int dpitch)
{
	const unsigned char *srcp_saved = srcp;
	const unsigned char *srcpp;
	const unsigned char *srcpn;

	unsigned char *workp_saved = workp;
	unsigned char *workpp;
	unsigned char *workpn;

	unsigned char *maskp_saved = maskp;
	unsigned char *maskpp;
	unsigned char *maskpn;

	unsigned char *finalp_saved = finalp;

	int x, y, rsum, bsum, gsum;

	srcpp = srcp_saved;
 	srcp = srcpp + spitch;
	srcpn = srcp + spitch;
	maskpp = maskp_saved;
 	maskp = maskpp + dpitch;
	maskpn = maskp + dpitch;
	workp = workp_saved + dpitch;
	memcpy(workp_saved, srcp_saved, row_size);
	memcpy(workp_saved + (height-1)*dpitch, srcp_saved + (height-1)*spitch, row_size);
	for (y = 1; y < height - 1; y++)
	{
		workp[0] = srcp[0];
		workp[1] = srcp[1];
		workp[2] = srcp[2];
		workp[row_size-4] = srcp[row_size-4];
		workp[row_size-3] = srcp[row_size-3];
		workp[row_size-2] = srcp[row_size-2];
        for (x = 4; x < row_size - 4; x+=4)
		{
			int count = 1;

			if (!maskp[x])
			{
				bsum = srcp[x];
				gsum = srcp[x+1];
				rsum = srcp[x+2];
				if (!maskpp[x])
				{
					bsum += srcpp[x];
					gsum += srcpp[x+1];
					rsum += srcpp[x+2];
					count++;
				}
				if (!maskpn[x])
				{
					bsum += srcpn[x];
					gsum += srcpn[x+1];
					rsum += srcpn[x+2];
					count++;
				}

				workp[x]   = bsum / count;
				workp[x+1] = gsum / count;
				workp[x+2] = rsum / count;
			}
		}
        srcpp += spitch;
        srcp += spitch;
        srcpn += spitch;
 		maskpp += dpitch;
		maskp += dpitch;
		maskpn += dpitch;
        workp += dpitch;
    }
	srcpp = srcp_saved;
 	srcp = srcpp + spitch;
	srcpn = srcp + spitch;
	workpp = workp_saved;
 	workp = workpp + dpitch;
	workpn = workp + dpitch;
	maskpp = maskp_saved;
 	maskp = maskpp + dpitch;
	maskpn = maskp + dpitch;
	finalp = finalp_saved + dpitch;
	for (y = 1; y < height - 1; y++)
	{
        for (x = 4; x < row_size - 4; x+=4)
		{
			int count = 1;
			if (!maskp[x])
			{
				bsum = workp[x];
				gsum = workp[x+1];
				rsum = workp[x+2];
				if (!maskp[x-4])
				{
					bsum += workp[x-4];
					gsum += workp[x-3];
					rsum += workp[x-2];
					count++;
				}
				if (!maskp[x+4])
				{
					bsum += workp[x+4];
					gsum += workp[x+5];
					rsum += workp[x+6];
					count++;
				}

				finalp[x]   = bsum / count;
				finalp[x+1] = gsum / count;
				finalp[x+2] = rsum / count;
			}
			else
			{
				*((unsigned int *)(&finalp[x])) = *((unsigned int *)(&srcp[x]));
			}
		}
        srcpp += spitch;
        srcp += spitch;
        srcpn += spitch;
        workpp += dpitch;
        workp += dpitch;
        workpn += dpitch;
 		maskpp += dpitch;
		maskp += dpitch;
		maskpn += dpitch;
        finalp += dpitch;
    }
}

AVSValue __cdecl Create_Msmooth(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	/* If the strength is zero, don't bother doing any filtering. */
	if (args[3].AsInt() == 0) return (args[0].AsClip());
    return new Msmooth(args[0].AsClip(),
					args[1].AsBool(false),	// show mask
					args[2].AsInt(15),		// Msmooth detect threshold
					args[3].AsInt(3),		// blur strength
					args[4].AsBool(true),	// Msmooth detection high quality
					args[5].AsBool(false),	// chroma
					args[6].AsBool(false),	// show
					args[7].AsBool(false),	// debug
					env);
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
    env->AddFunction("Msmooth", "c[mask]b[threshold]i[strength]i[highq]b[chroma]b[show]b[debug]b", Create_Msmooth, 0);
    return 0;
}

