/*
	MSmooth plugin for Avisynth -- performs detail-preserving smoothing.

	Copyright (C) 2003 Donald A. Graft

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define VERSION "MSmooth 2.0.2"
#define COPYRIGHT "(c) Copyright 2003, Donald Graft"

class Msmooth : public GenericVideoFilter {
	bool showmask, highq, chroma, show, debug;
	int threshold, strength;

public:
    Msmooth(PClip _child, bool _showmask, int _threshold, int _strength, bool _highq, bool _chroma, bool _show,
		 bool _debug, IScriptEnvironment* env) :
		 GenericVideoFilter(_child), showmask(_showmask), threshold(_threshold), strength(_strength), highq(_highq),
		 chroma(_chroma), show(_show), debug(_debug)
	{
		if (!vi.IsRGB() && !vi.IsYV12())
			env->ThrowError("Msmooth: RGB32 or YV12 data only");
		if (strength > 25)
			env->ThrowError("Msmooth: strength must be 25 or less");
		if (debug)
		{
			char b[80];
			sprintf(b, "%s by Donald Graft, Copyright 2003\n", VERSION);
			OutputDebugString(b);
		}
	}
    PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
    PVideoFrame __stdcall GetFrameRGB(int n, IScriptEnvironment* env);
    PVideoFrame __stdcall GetFrameYV12(int n, IScriptEnvironment* env);
	void __stdcall SmoothingPassRGB(const unsigned char *srcp, unsigned char *maskp, unsigned char *workp, unsigned char *finalp,
		                         int row_size, int height, int spitch, int dpitch);
	void __stdcall EdgeMaskRGB(const unsigned char *srcp, unsigned char *blurp, unsigned char *workp, unsigned char *maskp,
		                         int row_size, int height, int src_pitch, int blur_pitch);
	void __stdcall SmoothingPassYV12(const unsigned char *srcp, unsigned char *maskp, unsigned char *workp, unsigned char *finalp,
		                         int row_size, int height, int spitch, int dpitch);
	void __stdcall EdgeMaskYV12(const unsigned char *srcp, unsigned char *blurp, unsigned char *workp, unsigned char *maskp,
		                         int row_size, int height, int src_pitch, int blur_pitch);
};
