/*
	MSmooth plugin for Avisynth -- performs detail-preserving smoothing.

	Copyright (C) 2003 Donald A. Graft

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "internal.h"
#include "windows.h"
#include "stdio.h"
#include "msmooth.h"
#include "info.h"

PVideoFrame __stdcall Msmooth::GetFrameYV12(int n, IScriptEnvironment* env)
{
    PVideoFrame src = child->GetFrame(n, env);
    PVideoFrame blur = env->NewVideoFrame(vi);
    PVideoFrame work = env->NewVideoFrame(vi);
    PVideoFrame mask = env->NewVideoFrame(vi);
    PVideoFrame final = env->NewVideoFrame(vi);
    PVideoFrame final2 = env->NewVideoFrame(vi);
	PVideoFrame deliver;

	const unsigned char *srcpY = src->GetReadPtr(PLANAR_Y);
	const unsigned char *srcp_savedY = srcpY;
    int src_pitchY = src->GetPitch(PLANAR_Y);
	const unsigned char *srcpU = src->GetReadPtr(PLANAR_U);
	const unsigned char *srcp_savedU = srcpU;
    int src_pitchUV = src->GetPitch(PLANAR_U);
	const unsigned char *srcpV = src->GetReadPtr(PLANAR_V);
	const unsigned char *srcp_savedV = srcpV;

    unsigned char *blurpY = blur->GetWritePtr(PLANAR_Y);
	unsigned char *blurp_savedY = blurpY;
    int blur_pitchY = blur->GetPitch(PLANAR_Y);
    unsigned char *blurpU = blur->GetWritePtr(PLANAR_U);
	unsigned char *blurp_savedU = blurpU;
    int blur_pitchUV = blur->GetPitch(PLANAR_U);
    unsigned char *blurpV = blur->GetWritePtr(PLANAR_V);
	unsigned char *blurp_savedV = blurpV;

    unsigned char *workpY = work->GetWritePtr(PLANAR_Y);
	unsigned char *workp_savedY = workpY;
    unsigned char *workpU = work->GetWritePtr(PLANAR_U);
	unsigned char *workp_savedU = workpU;
    unsigned char *workpV = work->GetWritePtr(PLANAR_V);
	unsigned char *workp_savedV = workpV;

    unsigned char *maskpY = mask->GetWritePtr(PLANAR_Y);
	unsigned char *maskp_savedY = maskpY;
    unsigned char *maskpUV = mask->GetWritePtr(PLANAR_U);
	unsigned char *maskp_savedUV = maskpUV;

    unsigned char *finalpY = final->GetWritePtr(PLANAR_Y);
	unsigned char *finalp_savedY = finalpY;
    unsigned char *finalpU = final->GetWritePtr(PLANAR_U);
	unsigned char *finalp_savedU = finalpU;
    unsigned char *finalpV = final->GetWritePtr(PLANAR_V);
	unsigned char *finalp_savedV = finalpV;

    unsigned char *finalp2Y = final2->GetWritePtr(PLANAR_Y);
	unsigned char *finalp2_savedY = finalp2Y;
    unsigned char *finalp2U = final2->GetWritePtr(PLANAR_U);
	unsigned char *finalp2_savedU = finalp2U;
    unsigned char *finalp2V = final2->GetWritePtr(PLANAR_V);
	unsigned char *finalp2_savedV = finalp2V;

    int row_sizeY = blur->GetRowSize(PLANAR_Y);
    int row_sizeUV = blur->GetRowSize(PLANAR_U);
    int heightY = blur->GetHeight(PLANAR_Y);
    int heightUV = blur->GetHeight(PLANAR_U);
	int x, y, reps;

	/* Create the detail mask. */
	EdgeMaskYV12(srcpY, blurpY, workpY, maskpY, row_sizeY, heightY, src_pitchY, blur_pitchY);

	/* Create a downsized version for the chroma smoothing. It will be used for
	   both U and V. */
	if (chroma)
	{
		maskpY = maskp_savedY;
		for (y = 0; y < heightUV; y++)
		{
			for (x = 0; x < row_sizeUV; x++)
			{
				maskpUV[x] = maskpY[2*x];
			}
			maskpY += 2*blur_pitchY;
			maskpUV += blur_pitchUV;
		}
	}

	if (showmask == true)
	{
		if (show == true)
		{
			char buf[80];
			env->MakeWritable(&mask);
			sprintf(buf, VERSION);
			DrawStringYV12(mask, 0, 0, buf);
			sprintf(buf, COPYRIGHT);
			DrawStringYV12(mask, 0, 1, buf);
		}
		return mask;
	}

	/* Fix up output frame borders. */
	srcpY = srcp_savedY;
	finalpY = finalp_savedY;
	finalp2Y = finalp2_savedY;
	memcpy(finalpY, srcpY, row_sizeY);
	memcpy(finalpY + (heightY-1)*blur_pitchY, srcpY + (heightY-1)*blur_pitchY, row_sizeY);
	memcpy(finalp2Y, srcpY, row_sizeY);
	memcpy(finalp2Y + (heightY-1)*blur_pitchY, srcpY + (heightY-1)*blur_pitchY, row_sizeY);
	for (y = 0; y < heightY; y++)
	{
		finalpY[0] = finalp2Y[0] = srcpY[0];
		finalpY[row_sizeY-1] = finalp2Y[row_sizeY-1] = srcpY[row_sizeY-1];
		srcpY += src_pitchY;
		finalpY += blur_pitchY;
		finalp2Y += blur_pitchY;
	}
	srcpU = srcp_savedU;
	finalpU = finalp_savedU;
	finalp2U = finalp2_savedU;
	memcpy(finalpU, srcpU, row_sizeUV);
	memcpy(finalpU + (heightUV-1)*blur_pitchUV, srcpU + (heightUV-1)*blur_pitchUV, row_sizeUV);
	memcpy(finalp2U, srcpU, row_sizeUV);
	memcpy(finalp2U + (heightUV-1)*blur_pitchUV, srcpU + (heightUV-1)*blur_pitchUV, row_sizeUV);
	for (y = 0; y < heightUV; y++)
	{
		finalpU[0] = finalp2U[0] = srcpU[0];
		finalpU[row_sizeUV-1] = finalp2U[row_sizeUV-1] = srcpU[row_sizeUV-1];
		srcpU += src_pitchUV;
		finalpU += blur_pitchUV;
		finalp2U += blur_pitchUV;
	}
	srcpV = srcp_savedV;
	finalpV = finalp_savedV;
	finalp2V = finalp2_savedV;
	memcpy(finalpV, srcpV, row_sizeUV);
	memcpy(finalpV + (heightUV-1)*blur_pitchUV, srcpV + (heightUV-1)*blur_pitchUV, row_sizeUV);
	memcpy(finalp2V, srcpV, row_sizeUV);
	memcpy(finalp2V + (heightUV-1)*blur_pitchUV, srcpV + (heightUV-1)*blur_pitchUV, row_sizeUV);
	for (y = 0; y < heightUV; y++)
	{
		finalpV[0] = finalp2V[0] = srcpV[0];
		finalpV[row_sizeUV-1] = finalp2V[row_sizeUV-1] = srcpV[row_sizeUV-1];
		srcpV += src_pitchUV;
		finalpV += blur_pitchUV;
		finalp2V += blur_pitchUV;
	}

	/* Masked smoothing. */
	reps = 0;
	SmoothingPassYV12(srcp_savedY, maskp_savedY, workp_savedY, finalp_savedY, row_sizeY, heightY, src_pitchY, blur_pitchY);
	if (chroma)
	{
		SmoothingPassYV12(srcp_savedU, maskp_savedUV, workp_savedU, finalp_savedU, row_sizeUV, heightUV, src_pitchUV, blur_pitchUV);
		SmoothingPassYV12(srcp_savedV, maskp_savedUV, workp_savedV, finalp_savedV, row_sizeUV, heightUV, src_pitchUV, blur_pitchUV);
	}
	if (++reps >= strength)
	{
		deliver = final;
		goto done;
	}
	while (1)
	{
		SmoothingPassYV12(finalp_savedY, maskp_savedY, workp_savedY, finalp2_savedY, row_sizeY, heightY, blur_pitchY, blur_pitchY);
		if (chroma)
		{
			SmoothingPassYV12(finalp_savedU, maskp_savedUV, workp_savedU, finalp2_savedU, row_sizeUV, heightUV, blur_pitchUV, blur_pitchUV);
			SmoothingPassYV12(finalp_savedV, maskp_savedUV, workp_savedV, finalp2_savedV, row_sizeUV, heightUV, blur_pitchUV, blur_pitchUV);
		}
		if (++reps >= strength)
		{
			deliver = final2;
			goto done;
		}
		SmoothingPassYV12(finalp2_savedY, maskp_savedY, workp_savedY, finalp_savedY, row_sizeY, heightY, blur_pitchY, blur_pitchY);
		if (chroma)
		{
			SmoothingPassYV12(finalp2_savedU, maskp_savedUV, workp_savedU, finalp_savedU, row_sizeUV, heightUV, blur_pitchUV, blur_pitchUV);
			SmoothingPassYV12(finalp2_savedV, maskp_savedUV, workp_savedV, finalp_savedV, row_sizeUV, heightUV, blur_pitchUV, blur_pitchUV);
		}
		if (++reps >= strength)
		{
			deliver = final;
			goto done;
		}
	}

done:
	if (show == true)
	{
		char buf[80];
		env->MakeWritable(&deliver);
		sprintf(buf, VERSION);
		DrawStringYV12(deliver, 0, 0, buf);
		sprintf(buf, COPYRIGHT);
		DrawStringYV12(deliver, 0, 1, buf);
	}
	if (!chroma)
	{
		srcpU = srcp_savedU;
		srcpV = srcp_savedV;
		finalpU = deliver->GetWritePtr(PLANAR_U);
		finalpV = deliver->GetWritePtr(PLANAR_V);
		for (y = 0; y < heightUV; y++)
		{
			memcpy(finalpU, srcpU, row_sizeUV);
			memcpy(finalpV, srcpV, row_sizeUV);
			srcpU += src_pitchUV;
			srcpV += src_pitchUV;
			finalpU += blur_pitchUV;
			finalpV += blur_pitchUV;
		}
	}

	return(deliver);
}

void __stdcall Msmooth::EdgeMaskYV12(const unsigned char *srcp, unsigned char *blurp, unsigned char *workp, unsigned char *maskp,
                        int row_size, int height, int src_pitch, int blur_pitch)
{
	const unsigned char *srcp_saved = srcp;
	const unsigned char *srcpp;
	const unsigned char *srcpn;

	unsigned char *blurp_saved = blurp;
	const unsigned char *blurpn;

	unsigned char *workp_saved = workp;
	unsigned char *workpp;
	unsigned char *workpn;

	unsigned char *maskp_saved = maskp;

	int x, y;
	int y1, y2, y3, y4;

	/* Blur the source image prior to detail detection. */
	srcpp = srcp_saved;
	srcp  = srcpp + src_pitch;
	srcpn = srcp + src_pitch;
	workp = workp_saved + blur_pitch;
	memcpy(workp_saved, srcp_saved, row_size);
	memcpy(workp_saved + (height-1)*blur_pitch, srcp_saved + (height-1)*src_pitch, row_size);
	for (y = 1; y < height - 1; y++)
	{
		workp[0] = srcp[0];
		workp[row_size-1] = srcp[row_size-1];
        for (x = 1; x < row_size - 1; x++)
		{
			workp[x] = (srcpp[x] + srcp[x] + srcpn[x]) / 3;
		}
        srcpp += src_pitch;
        srcp += src_pitch;
        srcpn += src_pitch;
        workp += blur_pitch;
    }

	srcp  = srcp_saved + src_pitch;
	workpp = workp_saved;
	workp  = workpp + blur_pitch;
	workpn = workp + blur_pitch;
	blurp  = blurp_saved + blur_pitch;
	memcpy(blurp_saved, srcp_saved, row_size);
	memcpy(blurp_saved + (height-1)*blur_pitch, srcp_saved + (height-1)*src_pitch, row_size);
	for (y = 1; y < height - 1; y++)
	{
		blurp[0] = srcp[0];
		blurp[row_size-1] = srcp[row_size-1];
        for (x = 1; x < row_size - 1; x++)
		{
			blurp[x] = (workp[x-1] + workp[x] + workp[x+1]) / 3;
		}
        srcp += src_pitch;
        workpp += blur_pitch;
        workp += blur_pitch;
        workpn += blur_pitch;
        blurp += blur_pitch;
    }

	/* Diagonal detail detection. */
	blurp = (unsigned char *) workp_saved;
	blurpn = blurp + blur_pitch;
	maskp = maskp_saved;
	y1 = blurp[0];
	y3 = blurpn[0];
	for (y = 0; y < height - 1; y++) 
	{
		for (x = 0; x < row_size - 1; x++)
		{
			y2 = blurp[x+1];
			y4 = blurpn[x+1];
			if ((abs(y1 - y4) >= threshold) || (abs(y2 - y3) >= threshold))
			{
				maskp[x] = 0xff;
			}
			else
			{
				maskp[x] = 0x0;
			}
			y1 = y2; y3 = y4;
		}
		maskp += blur_pitch;
		blurp += blur_pitch;
		blurpn += blur_pitch;
	}

	/* If set for high quality, also do horizontal and vertical
	   detail detection. */
	if (highq == true)
	{
		/* Msmooth convolve vertical. */
		for (x = 0; x < row_size; x++)
		{
 			blurp = blurp_saved;
			blurpn = blurp + blur_pitch;
			maskp = maskp_saved;
			y1 = blurp[x];
			for (y = 0; y < height - 1; y++)
			{
				y2 = blurpn[x];
				if (abs(y1 - y2) >= threshold)
				{
					maskp[x] = 0xff;
				}
				y1 = y2;
				maskp += blur_pitch;
				blurp += blur_pitch;
				blurpn += blur_pitch;
			}
		}

		/* Msmooth convolve horizontal. */
		blurp = blurp_saved;
		maskp = maskp_saved;
		for (y = 0; y < height; y++)
		{
			y1 = blurp[0];
			for (x = 0; x < row_size - 1; x++)
			{
				y2 = blurp[x+1];
				if (abs(y1 - y2) >= threshold)
				{
					maskp[x] = 0xff;
				}
				y1 = y2;
			}
			maskp += blur_pitch;
			blurp += blur_pitch;
		}

	}
	/* Fix up detail map borders. */
	maskp = maskp_saved;
	memset(maskp, 0xff, row_size);
	memset(maskp + (height-1)*blur_pitch, 0xff, row_size);
	for (y = 0; y < height; y++)
	{
//		*((unsigned int *)(&maskp[0])) = 0xffffffff;
//		*((unsigned int *)(&maskp[row_size-1])) = 0xffffffff;
		maskp[0] = 0xff;
		maskp[row_size-1] = 0xff;
		maskp += blur_pitch;
	}
}

void __stdcall Msmooth::SmoothingPassYV12(const unsigned char *srcp, unsigned char *maskp,
									  unsigned char *workp, unsigned char *finalp,
									  int row_size, int height, int spitch, int dpitch)
{
	const unsigned char *srcp_saved = srcp;
	const unsigned char *srcpp;
	const unsigned char *srcpn;

	unsigned char *workp_saved = workp;
	unsigned char *workpp;
	unsigned char *workpn;

	unsigned char *maskp_saved = maskp;
	unsigned char *maskpp;
	unsigned char *maskpn;

	unsigned char *finalp_saved = finalp;

	int x, y, ysum;

	srcpp = srcp_saved;
 	srcp = srcpp + spitch;
	srcpn = srcp + spitch;
	maskpp = maskp_saved;
 	maskp = maskpp + dpitch;
	maskpn = maskp + dpitch;
	workp = workp_saved + dpitch;
	memcpy(workp_saved, srcp_saved, row_size);
	memcpy(workp_saved + (height-1)*dpitch, srcp_saved + (height-1)*spitch, row_size);
	for (y = 1; y < height - 1; y++)
	{
		workp[0] = srcp[0];
		workp[row_size-1] = srcp[row_size-1];
        for (x = 1; x < row_size - 1; x++)
		{
			int count = 1;

			if (!maskp[x])
			{
				ysum = srcp[x];
				if (!maskpp[x])
				{
					ysum += srcpp[x];
					count++;
				}
				if (!maskpn[x])
				{
					ysum += srcpn[x];
					count++;
				}
				workp[x] = ysum / count;
			}
		}
        srcpp += spitch;
        srcp += spitch;
        srcpn += spitch;
 		maskpp += dpitch;
		maskp += dpitch;
		maskpn += dpitch;
        workp += dpitch;
    }
	srcpp = srcp_saved;
 	srcp = srcpp + spitch;
	srcpn = srcp + spitch;
	workpp = workp_saved;
 	workp = workpp + dpitch;
	workpn = workp + dpitch;
	maskpp = maskp_saved;
 	maskp = maskpp + dpitch;
	maskpn = maskp + dpitch;
	finalp = finalp_saved + dpitch;
	for (y = 1; y < height - 1; y++)
	{
        for (x = 1; x < row_size - 1; x++)
		{
			int count = 1;

			if (!maskp[x])
			{
				ysum = workp[x];
				if (!maskp[x-1])
				{
					ysum += workp[x-1];
					count++;
				}
				if (!maskp[x+1])
				{
					ysum += workp[x+1];
					count++;
				}
				finalp[x] = ysum / count;
			}
			else
			{
				finalp[x] = srcp[x];
			}
		}
        srcpp += spitch;
        srcp += spitch;
        srcpn += spitch;
        workpp += dpitch;
        workp += dpitch;
        workpn += dpitch;
 		maskpp += dpitch;
		maskp += dpitch;
		maskpn += dpitch;
        finalp += dpitch;
    }
}
